/**************************************************************************
                                METRICS
   Copyright (C) 2014 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   CUTASTRE is an accelerated variant of ASTRE for the detection of
   smooth trajectories in a (noisy) point set sequence. This variant
   manages to break the quadratic time and memory complexity of ASTRE
   with respect to the number of frames, while showing similar (or even
   slightly better) detection performances. The resulting algorithm
   CUTASTRE has a linear complexity with respect to the number of
   frames.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation : gcc -O3 metrics.c -lm -o metrics

  References :

  + [1] ASTRE    : Maël Primet and Lionel Moisan, ``Point tracking: an a-contrario approach'', preprint MAP5 nº2012-06.
                   http://www.math-info.univ-paris5.fr/~moisan/papers/2012-06.pdf
  + [2] CUTASTRE : Rémy Abergel and Lionel Moisan, ``Accelerated A-contrario Detection of Smooth Trajectories'', Proceedings of EUSIPCO (2014).
                   http://www.eurasip.org/Proceedings/Eusipco/Eusipco2014/HTML/papers/1569926799.pdf

  Websites :

    + ASTRE : http://www.math-info.univ-paris5.fr/~moisan/astre/
    + CUTASTRE (under construction) : http://www.math-info.univ-paris5.fr/~moisan/cutastre/

  Versions :

  + v 1.0 : (08/2015) initial version (Rémy Abergel)

 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define VERSION "1.0"

/* basic macros */
#define MAX(a,b)        (((a)>(b)) ? (a) : (b))

/* point structures */
typedef struct point {
  int frameid,x,y,trajid,next,posinref;
} Point;


/* load the points of a sequence, size = number of points loaded */
static void retrieve_sequence(filename,ptseq,size)
     char *filename;
     Point **ptseq;
     int *size;
{
  FILE *file;
  char *line;
  int e,n,c,len=0,lenMax=0,nlines=0,ncomments=0;

  /* open input file, count number of (nonempty) lines and compute the maximal length of a line */
  if(NULL == (file = fopen(filename,"r"))) {
    printf("Fatal error: could not open file '%s'.\n",filename);
    exit(EXIT_FAILURE);
  }
  while((c = fgetc(file)) != EOF) {
    len++;
    if(c == '\n') { nlines++; lenMax = MAX(len,lenMax); len=0; }
  }
  rewind(file);
  lenMax++;
  if(NULL == (line = (char*) malloc (lenMax*sizeof(char)))) {
    printf("Fatal error: Not enough memory.\n");
    fclose(file); exit(EXIT_FAILURE);
  }

  /* count the number of lines starting with character '#' (comments) */
  for (n=0;n<nlines;n++) {
    if(NULL == fgets(line,lenMax,file)) {
      printf("An error occured while reading input file.\n");
      fclose(file); free(line); exit(EXIT_FAILURE);
    }
    if(line[0]=='#') ncomments++;
  }
  rewind(file);

  /* allocate memory for the point sequence and retrieve data */
  *size = nlines - ncomments;
  if(NULL == (*ptseq = (Point*) malloc ((*size)*sizeof(Point)))) {
    printf("Fatal error: Not enough memory.\n");
    fclose(file); free(line); exit(EXIT_FAILURE);
  }
  for(n=c=0;c<nlines;c++) {
    if(NULL == fgets(line,lenMax,file)) {
      printf("An error occured while reading input file.\n");
      fclose(file); free(line); free(*ptseq); exit(EXIT_FAILURE);
    }
    if(line[0] != '#') {
      if (n < *size) {
	e = sscanf(line,"%d %d %d %d",&((*ptseq)[n].frameid),&((*ptseq)[n].x),&((*ptseq)[n].y),&((*ptseq)[n].trajid));
	(*ptseq)[n].next = (*ptseq)[n].posinref = -1;
      }
      if ((e != 4)||(n>=(*size))) {
	printf("Fatal error: could not correctly retrieve data from the input point sequence.\n");
	printf("Please check the format of your input point set sequence, each line must\nstart with four integers ``<frameid> <x> <y> <trajid>'' (separated with space) OR\nstart with character '#' (comment). Notice that empty lines are not allowed!\n");
	fclose(file); free(line); free(*ptseq); exit(EXIT_FAILURE);
      }
      n++;
    }
  }
  fclose(file); free(line); line = NULL;
}


/* set nlinks_ref and nlinks_seq = number of links in ref and seq, set ncorrect_links = number of correct links in ref */
static void process_data(ref,seq,size,nlinks_ref,nlinks_seq,ncorrect_links)
     Point *ref,*seq;
     int size,*nlinks_ref,*nlinks_seq,*ncorrect_links;
{
  int i,j,idmax,*idcount;
  Point *pt1,*pt2;
  char found;

  /* for each point of seq, retrieve position in ref */
  for(i=0;i<size;i++) {
    pt1 = &(seq[i]);
    found = j = 0;
    while((j<size)&&(!found)) {
      pt2 = &(ref[j]);
      found = (((*pt2).frameid == (*pt1).frameid) && ((*pt2).x == (*pt1).x) && ((*pt2).y == (*pt1).y));
      if (!found) j++;
    }
    if(!found) {
      printf("Fatal error: the two input sequences do not contain the same points!\n");
      free(ref); free(seq); exit(EXIT_FAILURE);
    }
    else (*pt1).posinref = j;
  }

  /* count links in ref (no need to retrieve links) */
  for(idmax=-1,i=0;i<size;i++) idmax = MAX(idmax,ref[i].trajid);
  if (NULL==(idcount = (int*) calloc((idmax+1),sizeof(int)))) {
    printf("Fatal error: Not enough memory!\n");
    free(ref); free(seq);
    exit(EXIT_FAILURE);
  }
  // now set idcount[i] = number of points having trajid = i
  for(j=0;j<size;j++){
    i = ref[j].trajid;
    if(i >= 0) idcount[i]++;
  }
  // compute the number of links in ref (number of links for trajectory i is idcount[i]-1)
  for(*nlinks_ref=0,i=0;i<=idmax;i++)
    *nlinks_ref += MAX(0,idcount[i]-1);
  free(idcount);

  /* retrieve links in seq */
  for(*ncorrect_links=*nlinks_seq=i=0;i<size;i++) {
    pt1 = &(seq[i]);
    found = ((*pt1).trajid == -1);
    j = 0;
    while((!found)&&(j<size)) {
      pt2 = &(seq[j]);
      if(((*pt2).frameid > ((*pt1).frameid))&&((*pt2).trajid == (*pt1).trajid)) {
	if((*pt1).next == -1) {
	  (*pt1).next = j;
	  *nlinks_seq += 1;
	}
	else if(((*pt2).frameid-(*pt1).frameid) < (seq[(*pt1).next].frameid-(*pt1).frameid))
	  (*pt1).next = j;
	found = ((*pt2).frameid == 1+(*pt1).frameid);
      }
      j++;
    }
    if((*pt1).next != -1) {
      pt2 = &seq[(*pt1).next];
      *ncorrect_links += (int) (ref[(*pt1).posinref].trajid == ref[(*pt2).posinref].trajid);
    }
  }

}


static void display_usage()
{
  printf("\nUsage: metrics [--help] [--version] ref seq\n\n");
  printf("   --help    : display help\n");
  printf("   --version : display program version\n");
  printf("   ref       : path to the reference point set sequence file\n");
  printf("   seq       : path to the tested point set sequence file\n\n");
}



int main(int argc, char **argv)
{
  char *path_ref = NULL,*path_seq = NULL;
  Point *ref,*seq;
  int i,size_ref,size_seq,nlinks_ref,nlinks_seq,ncorrect_links;
  double recall,precision,f1score,gmeasure;

  /******************************************/
  /*            arguments parser            */
  /******************************************/
  for(i=1;i<argc;i++) {
    if ((strcmp(argv[i],"--help") == 0) || (strcmp(argv[i],"-?") == 0)) {
      display_usage();
      printf("Input files: the input point sequences of this program must have the same\nformat as output sequences returned by ASTRE and CUTASTRE, they are simple\ntext files where each line describes a point of the sequence (expecting\ncomments lines starting with character '#'). More precisely, each line\nmust contain at least four items ``<frameid> <x> <y> <trajid>'' (separated\nwith space, any content following those 4 items will be ignored) OR start\nwith character '#' (comment). Notice that empty lines are not allowed!\n\n");
      printf("   + <frameid> : identifier of the frame containing the point (frames are\n                 numeroted from 0)\n");
      printf("   + <x>       : X-coordinate (horizontal axis) of the point\n");
      printf("   + <y>       : Y-coordinate (vertical axis) of the point\n");
      printf("   + <trajid>  : identifier of the trajectory to which the point belongs,\n                 the detected trajectories are numeroted from 0, while the\n                 spurious points (noise) have trajid = -1.\n\n");
      return EXIT_SUCCESS;
    }
    else if (strcmp(argv[i],"--version") == 0) {
      printf("\n*****************************************************************************\n                             METRICS version %s                             \n*****************************************************************************\n\nCopyright (C) 2014 Rémy Abergel (remy.abergel AT gmail.com) and Lionel Moisan\n(Lionel.Moisan AT parisdescartes.fr) Université Paris Descartes, Laboratoire\nMAP5, CNRS UMR 8145.\n\nThis program is free software: you can redistribute it and/or modify it under\nthe terms of the GNU General Public License as published by the Free Software\nFoundation, either version 3 of the License, or (at your option) any later\nversion.\n\nThis program is distributed in the hope that it will be useful, but WITHOUT\nANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\nFOR A PARTICULAR PURPOSE. See the GNU General Public License for more\ndetails.\n\nYou should have received a copy of the GNU General Public License along with\nthis program. If not, see <http://www.gnu.org/licenses/>.\n\n",VERSION);
      return EXIT_SUCCESS;
    }
    else if (NULL == path_ref) path_ref = argv[i];
    else if (NULL == path_seq) path_seq = argv[i];
    else { printf("\nCould not retrieve properly input arguments.\n"); display_usage(); return EXIT_FAILURE; }
  }
  if ((NULL == path_ref) || (NULL == path_seq)) {
    if (NULL == path_ref) printf("\nerror: reference point sequence (ref) missing\n");
    else if(NULL == path_seq) printf("\nerror: tested point sequence (seq) missing\n");
    display_usage();
    return EXIT_FAILURE;
  }

  /****************************************************/
  /*                      KERNEL                      */
  /****************************************************/

  /* retrieve input data */
  retrieve_sequence(path_ref,&ref,&size_ref);
  retrieve_sequence(path_seq,&seq,&size_seq);
  if(size_ref != size_seq) {
    printf("Fatal error: the two input sequences do not contain the same points!\n");
    free(ref); free(seq); exit(EXIT_FAILURE);
  }

  /* process data (retrieve links, count number of correct links in seq) */
  process_data(ref,seq,size_ref,&nlinks_ref,&nlinks_seq,&ncorrect_links);

  /* compute & display measurements */
  precision = (double)ncorrect_links/(double)nlinks_ref;
  recall = (double)ncorrect_links/(double)nlinks_seq;
  f1score = 2*recall*precision/(recall+precision); // harmonic mean
  gmeasure = sqrt(precision*recall); // geometric mean
  printf("----------------------------------------------------------------------\n");
  printf("number of links found into the reference sequence      : %d\n",nlinks_ref);
  printf("number of links found into the tested sequence         : %d\n",nlinks_seq);
  printf("number of correct links found into the tested sequence : %d\n",ncorrect_links);
  printf("precision = %.10g\n",precision);
  printf("recall    = %.10g\n",recall);
  printf("F1-score  = %.10g\n",f1score);
  printf("G-measure = %.10g\n",gmeasure);
  printf("----------------------------------------------------------------------\n");

  /* free memory */
  free(ref);
  free(seq);

  return EXIT_SUCCESS;
}
